@@ -14,12 +14,14 @@
 #    under the License.
 
 import logging
+import os
 import queue
 import threading
 import time
 import uuid
 
 import cachetools
+from oslo_concurrency import lockutils
 from oslo_utils import eventletutils
 from oslo_utils import timeutils
 
@@ -40,6 +42,63 @@ ACK_REQUEUE_EVERY_SECONDS_MIN = 0.001
 ACK_REQUEUE_EVERY_SECONDS_MAX = 5.0
 
 
+class QManager(object):
+    """Queue Manager to build queue name for reply (and fanout) type.
+    This class is used only when use_queue_manager is set to True in config
+    file.
+    It rely on a shared memory accross processes (reading/writing data to
+    /dev/shm/xyz) and oslo_concurrency.lockutils to avoid assigning the same
+    queue name twice (or more) to different processes.
+    The original idea of this queue manager was to avoid random queue names,
+    so on service restart, the previously created queues can be reused,
+    avoiding deletion/creation of queues on rabbitmq side (which cost a lot
+    at scale).
+    """
+    def __init__(self, hostname, processname):
+        # We will use hostname and processname in queue names to help identify
+        # them easily.
+        # This is also ensuring consistency between service restart.
+        self.hostname = hostname
+        self.processname = processname
+        # This is where the counter is kept
+        self.file_name = '/dev/shm/%s_%s_qmanager' % (self.hostname,  # nosec
+                                                      self.processname)
+        # We use the process group to restart the counter on service restart
+        self.pg = os.getpgrp()
+
+    def get(self):
+        lock_name = 'oslo_read_shm_%s_%s' % (self.hostname, self.processname)
+
+        @lockutils.synchronized(lock_name, external=True)
+        def read_from_shm():
+            # Grab the counter from shm
+            # This function is thread and process safe thanks to lockutils
+            try:
+                with open(self.file_name, 'r') as f:
+                    pg, c = f.readline().split(':')
+                    pg = int(pg)
+                    c = int(c)
+            except (FileNotFoundError, ValueError):
+                pg = self.pg
+                c = 0
+
+            # Increment the counter
+            if pg == self.pg:
+                c += 1
+            else:
+                # The process group changed, maybe service restarted?
+                # Start over the counter
+                c = 1
+
+            # Write the new counter
+            with open(self.file_name, 'w') as f:
+                f.write(str(self.pg) + ':' + str(c))
+            return c
+
+        c = read_from_shm()
+        return self.hostname + ":" + self.processname + ":" + str(c)
+
+
 class MessageOperationsHandler(object):
     """Queue used by message operations to ensure that all tasks are
     serialized and run in the same thread, since underlying drivers like kombu
@@ -89,11 +148,11 @@ class AMQPIncomingMessage(base.RpcIncomingMessage):
     def __init__(self, listener, ctxt, message, unique_id, msg_id, reply_q,
                  client_timeout, obsolete_reply_queues,
                  message_operations_handler):
-        super(AMQPIncomingMessage, self).__init__(ctxt, message)
+        super(AMQPIncomingMessage, self).__init__(ctxt, message, msg_id)
+        self.orig_msg_id = msg_id
         self.listener = listener
 
         self.unique_id = unique_id
-        self.msg_id = msg_id
         self.reply_q = reply_q
         self.client_timeout = client_timeout
         self._obsolete_reply_queues = obsolete_reply_queues
@@ -125,7 +184,7 @@ class AMQPIncomingMessage(base.RpcIncomingMessage):
         conn.direct_send(self.reply_q, rpc_common.serialize_msg(msg))
 
     def reply(self, reply=None, failure=None):
-        if not self.msg_id:
+        if not self.orig_msg_id:
             # NOTE(Alexei_987) not sending reply, if msg_id is empty
             #    because reply should not be expected by caller side
             return
@@ -437,14 +496,21 @@ class ReplyWaiters(object):
         self._wrn_threshold = 10
 
     def get(self, msg_id, timeout):
-        try:
-            return self._queues[msg_id].get(block=True, timeout=timeout)
-        except queue.Empty:
-            raise oslo_messaging.MessagingTimeout(
-                'Timed out waiting for a reply '
-                'to message ID %s' % msg_id)
+        watch = timeutils.StopWatch(duration=timeout)
+        watch.start()
+        while not watch.expired():
+            try:
+                # NOTE(amorin) we can't use block=True
+                # See lp-2035113
+                return self._queues[msg_id].get(block=False)
+            except queue.Empty:
+                time.sleep(0.5)
+        raise oslo_messaging.MessagingTimeout(
+            'Timed out waiting for a reply '
+            'to message ID %s' % msg_id)
 
     def put(self, msg_id, message_data):
+        LOG.debug('Received RPC response for msg %s', msg_id)
         queue = self._queues.get(msg_id)
         if not queue:
             LOG.info('No calling threads waiting for msg_id : %s', msg_id)
@@ -523,9 +589,11 @@ class ReplyWaiter(object):
         self.waiters.remove(msg_id)
 
     @staticmethod
-    def _raise_timeout_exception(msg_id):
+    def _raise_timeout_exception(msg_id, reply_q):
         raise oslo_messaging.MessagingTimeout(
-            'Timed out waiting for a reply to message ID %s.', msg_id)
+            'Timed out waiting for a reply %(reply_q)s '
+            'to message ID %(msg_id)s.',
+            {'msg_id': msg_id, 'reply_q': reply_q})
 
     def _process_reply(self, data):
         self.msg_id_cache.check_duplicate_message(data)
@@ -539,7 +607,7 @@ class ReplyWaiter(object):
         ending = data.get('ending', False)
         return result, ending
 
-    def wait(self, msg_id, timeout, call_monitor_timeout):
+    def wait(self, msg_id, timeout, call_monitor_timeout, reply_q):
         # NOTE(sileht): for each msg_id we receive two amqp message
         # first one with the payload, a second one to ensure the other
         # have finish to send the payload
@@ -557,16 +625,26 @@ class ReplyWaiter(object):
         final_reply = None
         ending = False
         while not ending:
-            timeout = timer.check_return(self._raise_timeout_exception, msg_id)
+            timeout = timer.check_return(
+                self._raise_timeout_exception,
+                msg_id,
+                reply_q
+            )
             if call_monitor_timer and timeout > 0:
                 cm_timeout = call_monitor_timer.check_return(
-                    self._raise_timeout_exception, msg_id)
+                    self._raise_timeout_exception,
+                    msg_id,
+                    reply_q
+                )
                 if cm_timeout < timeout:
                     timeout = cm_timeout
             try:
                 message = self.waiters.get(msg_id, timeout=timeout)
             except queue.Empty:
-                self._raise_timeout_exception(msg_id)
+                self._raise_timeout_exception(
+                    msg_id,
+                    reply_q
+                )
 
             reply, ending = self._process_reply(message)
             if reply is not None:
@@ -597,6 +675,12 @@ class AMQPDriverBase(base.BaseDriver):
         self._reply_q = None
         self._reply_q_conn = None
         self._waiter = None
+        if conf.oslo_messaging_rabbit.use_queue_manager:
+            self._q_manager = QManager(
+                hostname=conf.oslo_messaging_rabbit.hostname,
+                processname=conf.oslo_messaging_rabbit.processname)
+        else:
+            self._q_manager = None
 
     def _get_exchange(self, target):
         return target.exchange or self._default_exchange
@@ -608,10 +692,16 @@ class AMQPDriverBase(base.BaseDriver):
 
     def _get_reply_q(self):
         with self._reply_q_lock:
+            # NOTE(amorin) Re-use reply_q when it already exists
+            # This avoid creating too many queues on AMQP server (rabbit)
             if self._reply_q is not None:
                 return self._reply_q
 
-            reply_q = 'reply_' + uuid.uuid4().hex
+            if self._q_manager:
+                reply_q = 'reply_' + self._q_manager.get()
+            else:
+                reply_q = 'reply_' + uuid.uuid4().hex
+            LOG.debug('Creating reply queue: %s', reply_q)
 
             conn = self._get_connection(rpc_common.PURPOSE_LISTEN)
 
@@ -628,12 +718,21 @@ class AMQPDriverBase(base.BaseDriver):
               envelope=True, notify=False, retry=None, transport_options=None):
 
         msg = message
+        reply_q = None
+        if 'method' in msg:
+            LOG.debug('Calling RPC method %s on target %s', msg.get('method'),
+                      target.topic)
+        else:
+            LOG.debug('Sending message to topic %s', target.topic)
 
         if wait_for_reply:
+            reply_q = self._get_reply_q()
             msg_id = uuid.uuid4().hex
             msg.update({'_msg_id': msg_id})
-            msg.update({'_reply_q': self._get_reply_q()})
+            msg.update({'_reply_q': reply_q})
             msg.update({'_timeout': call_monitor_timeout})
+            LOG.debug('Expecting reply to msg %s in queue %s', msg_id,
+                      reply_q)
 
         rpc_amqp._add_unique_id(msg)
         unique_id = msg[rpc_amqp.UNIQUE_ID]
@@ -676,7 +775,7 @@ class AMQPDriverBase(base.BaseDriver):
 
             if wait_for_reply:
                 result = self._waiter.wait(msg_id, timeout,
-                                           call_monitor_timeout)
+                                           call_monitor_timeout, reply_q)
                 if isinstance(result, Exception):
                     raise result
                 return result
