@@ -93,7 +93,13 @@ rabbit_opts = [
                 'an exception will be raised.'),
     cfg.BoolOpt('heartbeat_in_pthread',
                 default=False,
-                help="Run the health check heartbeat thread "
+                deprecated_for_removal=True,
+                deprecated_reason='The option is related to Eventlet which '
+                                  'will be removed. In addition this has '
+                                  'never worked as expected with services '
+                                  'using eventlet for core service framework.',
+                help="(DEPRECATED) It is recommend not to use this option "
+                     "anymore. Run the health check heartbeat thread "
                      "through a native python thread by default. If this "
                      "option is equal to False then the health check "
                      "heartbeat will inherit the execution model "
@@ -165,9 +171,16 @@ rabbit_opts = [
                 'Raft consensus algorithm. It is available as of '
                 'RabbitMQ 3.8.0. If set this option will conflict with '
                 'the HA queues (``rabbit_ha_queues``) aka mirrored queues, '
-                'in other words the HA queues should be disabled, quorum '
-                'queues durable by default so the amqp_durable_queues '
-                'opion is ignored when this option enabled.'),
+                'in other words the HA queues should be disabled. '
+                'Quorum queues are also durable by default so the '
+                'amqp_durable_queues option is ignored when this option is '
+                'enabled.'),
+    cfg.BoolOpt('rabbit_transient_quorum_queue',
+                default=False,
+                help='Use quorum queues for transients queues in RabbitMQ. '
+                'Enabling this option will then make sure those queues are '
+                'also using quorum kind of rabbit queues, which are HA by '
+                'default.'),
     cfg.IntOpt('rabbit_quorum_delivery_limit',
                default=0,
                help='Each time a message is redelivered to a consumer, '
@@ -234,7 +247,26 @@ rabbit_opts = [
                 default=False,
                 help="Enable x-cancel-on-ha-failover flag so that "
                      "rabbitmq server will cancel and notify consumers"
-                     "when queue is down")
+                     "when queue is down"),
+    cfg.BoolOpt('use_queue_manager',
+                default=False,
+                help='Should we use consistant queue names or random ones'),
+    cfg.StrOpt('hostname',
+               sample_default='node1.example.com',
+               default=socket.gethostname(),
+               help='Hostname used by queue manager. Defaults to the value '
+               'returned by socket.gethostname().'),
+    cfg.StrOpt('processname',
+               default=os.path.basename(sys.argv[0]),
+               help='Process name used by queue manager'),
+    cfg.BoolOpt('rabbit_stream_fanout',
+                default=False,
+                help='Use stream queues in RabbitMQ (x-queue-type: stream). '
+                'Streams are a new persistent and replicated data structure '
+                '("queue type") in RabbitMQ which models an append-only log '
+                'with non-destructive consumer semantics. It is available '
+                'as of RabbitMQ 3.9.0. If set this option will replace all '
+                'fanout queues with only one stream queue.'),
 ]
 
 LOG = logging.getLogger(__name__)
@@ -242,7 +274,8 @@ LOG = logging.getLogger(__name__)
 
 def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
                          rabbit_quorum_queue,
-                         rabbit_quorum_queue_config):
+                         rabbit_quorum_queue_config,
+                         rabbit_stream_fanout):
     """Construct the arguments for declaring a queue.
 
     If the rabbit_ha_queues option is set, we try to declare a mirrored queue
@@ -279,11 +312,15 @@ def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
 
     rabbit_quorum_queue_config:
     Quorum queues provides three options to handle message poisoning
-    and limit the resources the qourum queue can use
+    and limit the resources the quorum queue can use
     x-delivery-limit number of times the queue will try to deliver
     a message before it decide to discard it
     x-max-in-memory-length, x-max-in-memory-bytes control the size
     of memory used by quorum queue
+
+    If the rabbit_stream_fanout option is set, fanout queues are going to use
+    stream instead of quorum queues. See here:
+      https://www.rabbitmq.com/streams.html
     """
     args = {}
 
@@ -310,6 +347,12 @@ def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
     if rabbit_queue_ttl > 0:
         args['x-expires'] = rabbit_queue_ttl * 1000
 
+    if rabbit_stream_fanout:
+        args = {'x-queue-type': 'stream'}
+        if rabbit_queue_ttl > 0:
+            # max-age is a string
+            args['x-max-age'] = f"{rabbit_queue_ttl}s"
+
     return args
 
 
@@ -336,7 +379,8 @@ class Consumer(object):
                  exchange_auto_delete, queue_auto_delete, callback,
                  nowait=False, rabbit_ha_queues=None, rabbit_queue_ttl=0,
                  enable_cancel_on_failover=False, rabbit_quorum_queue=False,
-                 rabbit_quorum_queue_config=QuorumMemConfig(0, 0, 0)):
+                 rabbit_quorum_queue_config=QuorumMemConfig(0, 0, 0),
+                 rabbit_stream_fanout=False):
         """Init the Consumer class with the exchange_name, routing_key,
         type, durable auto_delete
         """
@@ -352,7 +396,7 @@ class Consumer(object):
         rabbit_quorum_queue_config = rabbit_quorum_queue_config or {}
         self.queue_arguments = _get_queue_arguments(
             rabbit_ha_queues, rabbit_queue_ttl, rabbit_quorum_queue,
-            rabbit_quorum_queue_config)
+            rabbit_quorum_queue_config, rabbit_stream_fanout)
         self.queue = None
         self._declared_on = None
         self.exchange = kombu.entity.Exchange(
@@ -361,6 +405,8 @@ class Consumer(object):
             durable=self.durable,
             auto_delete=self.exchange_auto_delete)
         self.enable_cancel_on_failover = enable_cancel_on_failover
+        self.rabbit_stream_fanout = rabbit_stream_fanout
+        self.next_stream_offset = "last"
 
     def _declare_fallback(self, err, conn, consumer_arguments):
         """Fallback by declaring a non durable queue.
@@ -395,6 +441,13 @@ class Consumer(object):
             )
             self.queue.declare()
 
+    def reset_stream_offset(self):
+        if not self.rabbit_stream_fanout:
+            return
+        LOG.warn("Reset consumer for queue %s next offset was at %s.",
+                 self.queue_name, self.next_stream_offset)
+        self.next_stream_offset = "last"
+
     def declare(self, conn):
         """Re-declare the queue after a rabbit (re)connect."""
 
@@ -403,6 +456,10 @@ class Consumer(object):
             consumer_arguments = {
                 "x-cancel-on-ha-failover": True}
 
+        if self.rabbit_stream_fanout:
+            consumer_arguments = {
+                "x-stream-offset": self.next_stream_offset}
+
         self.queue = kombu.entity.Queue(
             name=self.queue_name,
             channel=conn.channel,
@@ -415,8 +472,13 @@ class Consumer(object):
         )
 
         try:
-            LOG.debug('[%s] Queue.declare: %s',
-                      conn.connection_id, self.queue_name)
+            if self.rabbit_stream_fanout:
+                LOG.info('[%s] Stream Queue.declare: %s after offset %s',
+                         conn.connection_id, self.queue_name,
+                         self.next_stream_offset)
+            else:
+                LOG.debug('[%s] Queue.declare: %s',
+                          conn.connection_id, self.queue_name)
             try:
                 self.queue.declare()
             except amqp_ex.PreconditionFailed as err:
@@ -426,8 +488,33 @@ class Consumer(object):
                 # from each others. RabbitMQ will reject the services
                 # that try to create it with a configuration that differ
                 # from the one used first.
-                LOG.warning(err)
+                LOG.warning('[%s] Queue %s could not be declared probably '
+                            'because of conflicting configurations: %s',
+                            conn.connection_id, self.queue_name, err)
                 self._declare_fallback(err, conn, consumer_arguments)
+            except amqp_ex.NotFound as ex:
+                # NOTE(viktor.krivak): This exception is raised when
+                # non-durable and non-ha queue is hosted on node that
+                # is currently unresponsive or down. Thus, it is
+                # not possible to redeclare the queue since its status
+                # is unknown, and it could disappear in
+                # a few moments. However, the queue could be explicitly
+                # deleted and the redeclaration will then succeed.
+                # Queue will be then scheduled on any of the
+                # running/responsive nodes.
+                # This fixes bug:
+                # https://bugs.launchpad.net/oslo.messaging/+bug/2068630
+
+                LOG.warning("Queue %s is stuck on unresponsive node. "
+                            "Trying to delete the queue and redeclare it "
+                            "again, Error info: %s", self.queue_name, ex)
+                try:
+                    self.queue.delete()
+                except Exception as in_ex:
+                    LOG.warning("During cleanup of stuck queue deletion "
+                                "another exception occurred: %s. Ignoring...",
+                                in_ex)
+                self.queue.declare()
 
         except conn.connection.channel_errors as exc:
             # NOTE(jrosenboom): This exception may be triggered by a race
@@ -536,6 +623,12 @@ class Consumer(object):
         Messages that are processed and ack'ed.
         """
 
+        offset = message.headers.get("x-stream-offset")
+        if offset is not None:
+            LOG.debug("Stream for %s current offset: %s", self.queue_name,
+                      offset)
+            self.next_stream_offset = offset + 1
+
         m2p = getattr(self.queue.channel, 'message_to_python', None)
         if m2p:
             message = m2p(message)
@@ -639,6 +732,9 @@ class Connection(object):
         self.rabbit_quorum_queue = driver_conf.rabbit_quorum_queue
         self.rabbit_quorum_queue_config = self._get_quorum_configurations(
             driver_conf)
+        self.rabbit_transient_quorum_queue = \
+            driver_conf.rabbit_transient_quorum_queue
+        self.rabbit_stream_fanout = driver_conf.rabbit_stream_fanout
         self.rabbit_transient_queues_ttl = \
             driver_conf.rabbit_transient_queues_ttl
         self.rabbit_qos_prefetch_count = driver_conf.rabbit_qos_prefetch_count
@@ -656,6 +752,18 @@ class Connection(object):
         self.heartbeat_in_pthread = driver_conf.heartbeat_in_pthread
         self.ssl_enforce_fips_mode = driver_conf.ssl_enforce_fips_mode
         self.enable_cancel_on_failover = driver_conf.enable_cancel_on_failover
+        self.use_queue_manager = driver_conf.use_queue_manager
+
+        if self.rabbit_stream_fanout and self.rabbit_qos_prefetch_count <= 0:
+            raise RuntimeError('Configuration Error: rabbit_stream_fanout '
+                               'need rabbit_qos_prefetch_count to be set to '
+                               'a value greater than 0.')
+
+        if (self.rabbit_stream_fanout and not
+                self.rabbit_transient_quorum_queue):
+            raise RuntimeError('Configuration Error: rabbit_stream_fanout '
+                               'need rabbit_transient_quorum_queue to be set '
+                               'to true.')
 
         if self.heartbeat_in_pthread:
             # NOTE(hberaud): Experimental: threading module is in use to run
@@ -835,6 +943,13 @@ class Connection(object):
             self.connection.port = 1234
             self._poll_timeout = 0.05
 
+        if self.use_queue_manager:
+            self._q_manager = amqpdriver.QManager(
+                hostname=driver_conf.hostname,
+                processname=driver_conf.processname)
+        else:
+            self._q_manager = None
+
     # FIXME(markmc): use oslo sslutils when it is available as a library
     _SSL_PROTOCOLS = {
         "tlsv1": ssl.PROTOCOL_TLSv1,
@@ -981,11 +1096,24 @@ class Connection(object):
             info = {'err_str': exc, 'sleep_time': interval}
             info.update(self._get_connection_info(conn_error=True))
 
+            if 'Basic.cancel' in str(exc):
+                # This branch allows for consumer offset reset
+                # in the unlikely case consumers are cancelled. This may
+                # happen, for example, when we delete the stream queue.
+                # We need to start consuming from "last" because the stream
+                # offset maybe reset.
+                LOG.warn('[%s] Basic.cancel received. '
+                         'Resetting consumers offsets to last.',
+                         self.connection_id)
+                for consumer in self._consumers:
+                    consumer.reset_stream_offset()
+
             if 'Socket closed' in str(exc):
                 LOG.error('[%(connection_id)s] AMQP server'
                           ' %(hostname)s:%(port)s closed'
                           ' the connection. Check login credentials:'
                           ' %(err_str)s', info)
+
             else:
                 LOG.error('[%(connection_id)s] AMQP server on '
                           '%(hostname)s:%(port)s is unreachable: '
@@ -1095,11 +1223,20 @@ class Connection(object):
         """Close/release this connection."""
         self._heartbeat_stop()
         if self.connection:
-            for consumer in filter(lambda c: c.type == 'fanout',
-                                   self._consumers):
-                LOG.debug('[connection close] Deleting fanout '
-                          'queue: %s ' % consumer.queue.name)
-                consumer.queue.delete()
+            # NOTE(jcosmao) Delete queue should be called only when queue name
+            # is randomized. When using streams, queue is shared between
+            # all consumers, thus deleting fanout queue will force all other
+            # consumers to disconnect/reconnect by throwing
+            # amqp.exceptions.ConsumerCancelled.
+            # When using QManager, queue name is consistent accross agent
+            # restart, so we don't need to delete it either. Deletion must be
+            # handled by expiration policy.
+            if not self.rabbit_stream_fanout and not self.use_queue_manager:
+                for consumer in filter(lambda c: c.type == 'fanout',
+                                       self._consumers):
+                    LOG.debug('[connection close] Deleting fanout '
+                              'queue: %s ' % consumer.queue.name)
+                    consumer.queue.delete()
             self._set_current_channel(None)
             self.connection.release()
             self.connection = None
@@ -1345,13 +1482,15 @@ class Connection(object):
             queue_name=topic,
             routing_key='',
             type='direct',
-            durable=False,
+            durable=self.rabbit_transient_quorum_queue,
             exchange_auto_delete=False,
             queue_auto_delete=False,
             callback=callback,
             rabbit_ha_queues=self.rabbit_ha_queues,
             rabbit_queue_ttl=self.rabbit_transient_queues_ttl,
-            enable_cancel_on_failover=self.enable_cancel_on_failover)
+            enable_cancel_on_failover=self.enable_cancel_on_failover,
+            rabbit_quorum_queue=self.rabbit_transient_quorum_queue,
+            rabbit_quorum_queue_config=self.rabbit_quorum_queue_config)
 
         self.declare_consumer(consumer)
 
@@ -1377,22 +1516,35 @@ class Connection(object):
     def declare_fanout_consumer(self, topic, callback):
         """Create a 'fanout' consumer."""
 
-        unique = uuid.uuid4().hex
         exchange_name = '%s_fanout' % topic
-        queue_name = '%s_fanout_%s' % (topic, unique)
+        if self.rabbit_stream_fanout:
+            queue_name = '%s_fanout' % topic
+        else:
+            if self._q_manager:
+                unique = self._q_manager.get()
+            else:
+                unique = uuid.uuid4().hex
+            queue_name = '%s_fanout_%s' % (topic, unique)
+        LOG.debug('Creating fanout queue: %s', queue_name)
+
+        is_durable = (self.rabbit_transient_quorum_queue or
+                      self.rabbit_stream_fanout)
 
         consumer = Consumer(
             exchange_name=exchange_name,
             queue_name=queue_name,
             routing_key=topic,
             type='fanout',
-            durable=False,
+            durable=is_durable,
             exchange_auto_delete=True,
             queue_auto_delete=False,
             callback=callback,
             rabbit_ha_queues=self.rabbit_ha_queues,
             rabbit_queue_ttl=self.rabbit_transient_queues_ttl,
-            enable_cancel_on_failover=self.enable_cancel_on_failover)
+            enable_cancel_on_failover=self.enable_cancel_on_failover,
+            rabbit_quorum_queue=self.rabbit_transient_quorum_queue,
+            rabbit_quorum_queue_config=self.rabbit_quorum_queue_config,
+            rabbit_stream_fanout=self.rabbit_stream_fanout)
 
         self.declare_consumer(consumer)
 
@@ -1499,7 +1651,8 @@ class Connection(object):
                     self.rabbit_ha_queues,
                     0,
                     self.rabbit_quorum_queue,
-                    self.rabbit_quorum_queue_config))
+                    self.rabbit_quorum_queue_config,
+                    False))
             log_info = {'key': routing_key, 'exchange': exchange}
             LOG.trace(
                 'Connection._publish_and_creates_default_queue: '
@@ -1538,11 +1691,12 @@ class Connection(object):
 
     def direct_send(self, msg_id, msg):
         """Send a 'direct' message."""
-        exchange = kombu.entity.Exchange(name='',  # using default exchange
-                                         type='direct',
-                                         durable=False,
-                                         auto_delete=True,
-                                         passive=True)
+        exchange = kombu.entity.Exchange(
+            name='',  # using default exchange
+            type='direct',
+            durable=self.rabbit_transient_quorum_queue,
+            auto_delete=True,
+            passive=True)
         options = oslo_messaging.TransportOptions(
             at_least_once=self.direct_mandatory_flag)
 
@@ -1569,10 +1723,11 @@ class Connection(object):
 
     def fanout_send(self, topic, msg, retry=None):
         """Send a 'fanout' message."""
-        exchange = kombu.entity.Exchange(name='%s_fanout' % topic,
-                                         type='fanout',
-                                         durable=False,
-                                         auto_delete=True)
+        exchange = kombu.entity.Exchange(
+            name='%s_fanout' % topic,
+            type='fanout',
+            durable=self.rabbit_transient_quorum_queue,
+            auto_delete=True)
 
         LOG.debug('Sending fanout to %s_fanout', topic)
         self._ensure_publishing(self._publish, exchange, msg, retry=retry)
